@@ -23,9 +23,12 @@ fn new_lifecycle_item(lc: Lifecycle) -> LifecycleItem {
 pub struct Lifecycle {
     /// unique id
     id: LifecycleId,
-    ecu: DltChar4,
+    pub ecu: DltChar4,
     /// contains the number of messages belonging to this lifecycle. `0` indicates that this lifecycle is not valid anymore, e.g. was merged into different one.
     pub nr_msgs: u32,
+    /// number of control request messages. They are counted additionaly to identify lifecycles that consists of only control request messages. See [Self::only_control_requests()].
+    /// # Note: control request messages are treated differently as their timestamp is from a different clock domain (usually the logging device)
+    pub nr_control_req_msgs: u32,
     /// contains the start time of this lifecycle. See [Self::end_time()] as well. During processing this start_time is adjusted.
     /// # Note:
     /// This is not the reception time of the first message but the calculated start time of that lifecycle.
@@ -35,6 +38,7 @@ pub struct Lifecycle {
     /// The real start time will be slightly earlier as there is a minimal buffering time that is not considered / unknown.
     ///
     pub start_time: u64, // start time in us.
+    initial_start_time: u64,
     max_timestamp_us: u64, // max. timestamp of the messages assigned to this lifecycle. Used to determine end_time()
 }
 
@@ -79,15 +83,30 @@ impl Lifecycle {
         return self.start_time + self.max_timestamp_us;
     }
 
+    /// returns whether this lifecycle contains only control request messages.
+    /// ### Note: the info is wrong on merged lifecycles (we want to get rid of them anyhow)
+    pub fn only_control_requests(&self) -> bool {
+        self.nr_control_req_msgs >= self.nr_msgs
+    }
+
     /// create a new lifecycle with the first msg passed as parameter
     pub fn new(msg: &mut DltMessage) -> Lifecycle {
         // println!("new lifecycle created by {:?}", msg);
+        let is_ctrl_request = msg.is_ctrl_request();
+        let timestamp_us = if is_ctrl_request {
+            0
+        } else {
+            msg.timestamp_us()
+        };
+
         let alc = Lifecycle {
             id: NEXT_LC_ID.fetch_add(1, std::sync::atomic::Ordering::Relaxed),
             ecu: msg.ecu.clone(),
             nr_msgs: 1,
-            start_time: msg.reception_time_us - msg.timestamp_us(),
-            max_timestamp_us: msg.timestamp_us(),
+            nr_control_req_msgs: if is_ctrl_request { 1 } else { 0 },
+            start_time: msg.reception_time_us - timestamp_us,
+            initial_start_time: msg.reception_time_us - timestamp_us,
+            max_timestamp_us: timestamp_us,
         };
         msg.lifecycle = alc.id;
         alc
@@ -100,12 +119,14 @@ impl Lifecycle {
     pub fn merge(&mut self, lc_to_merge: &mut Lifecycle) {
         assert_ne!(lc_to_merge.nr_msgs, 0);
         self.nr_msgs += lc_to_merge.nr_msgs;
+        self.nr_control_req_msgs += lc_to_merge.nr_control_req_msgs;
         lc_to_merge.nr_msgs = 0; // this indicates a merged lc
         if lc_to_merge.max_timestamp_us > self.max_timestamp_us {
             self.max_timestamp_us = lc_to_merge.max_timestamp_us;
         }
         if lc_to_merge.start_time < self.start_time {
             self.start_time = lc_to_merge.start_time;
+            self.initial_start_time = lc_to_merge.initial_start_time;
         }
         // we mark this in the merged lc as max_timestamp_dms <- id
         lc_to_merge.max_timestamp_us = self.id as u64;
@@ -152,6 +173,16 @@ impl Lifecycle {
     ///
     pub fn update(&mut self, msg: &mut DltMessage) -> Option<Lifecycle> {
         // check whether this msg belongs to the lifecycle:
+        // 0) ignore any CTRL REQUEST msgs:
+        if msg.is_ctrl_request() {
+            // we dont check any params but
+            // simply add to this one
+            msg.lifecycle = self.id;
+            self.nr_msgs += 1;
+            self.nr_control_req_msgs += 1;
+            return None;
+        }
+
         // 1) the calc start time needs to be no later than the current end time
         // println!("update: lifecycle triggered to LC:{:?}", &self);
         let msg_timestamp_us = msg.timestamp_us();
@@ -165,22 +196,24 @@ impl Lifecycle {
             }
 
             // does it move the start to earlier? (e.g. has a smaller buffering delay)
+            // todo this can as well be caused by clock drift. Need to add clock_drift detection/compensation.
+            // the clock drift is relative to the recording devices time clock.
             if msg_lc_start < self.start_time {
+                /*println!(
+                    "update: lc {} #{} moving lc start_time by {}us to {} initial diff {}us",
+                    self.id, self.nr_msgs, self.start_time - msg_lc_start, msg_lc_start, self.initial_start_time - msg_lc_start
+                );*/
                 self.start_time = msg_lc_start;
-                println!(
-                    "update: lc {} moving lc start_time to {}",
-                    self.id, self.start_time
-                );
             }
             msg.lifecycle = self.id;
             self.nr_msgs += 1;
             // println!("update: lifecycle updated by {:?} to LC:{:?}", msg, &self);
             None
         } else {
-            println!(
+            /*println!(
                 "update: new lifecycle created by {:?} as msg_lc_start {} > {} LC:{:?}",
                 msg, msg_lc_start, cur_end_time, &self
-            );
+            );*/
             // new lifecycle:
             Some(Lifecycle::new(msg))
         }
@@ -306,11 +339,11 @@ where
 
                     // have to refresh here as a new lc was created to ensure that lcs always contains all lcs referenced by the msgs
                     lcs_w.refresh();
-                    for a in lcs_w.read().iter() {
+                    /* for a in lcs_w.read().iter() {
                         for (id, b) in a {
                             println!("lcs_w content id={:?} lc={:?}", id, b);
                         }
-                    }
+                    }*/
                 }
             }
             if remove_last_lc {
@@ -329,6 +362,7 @@ where
         outflow.send(msg).unwrap(); // todo how handle errors?
     }
     lcs_w.refresh();
+    /*
     println!(
         "After processing stream: Have ecu_map.len={}",
         ecu_map.len()
@@ -342,7 +376,7 @@ where
         for (id, b) in a {
             println!("lcs_w content id={:?} lc={:?}", id, b);
         }
-    }
+    }*/
     lcs_w
 }
 
@@ -407,10 +441,11 @@ where
         }
     }
 
+    /*
     println!("Have ecu_map.len={}", ecu_map.len());
     for (k, v) in &ecu_map {
         println!("Have for ecu {:?} {:?}", &k, &v);
-    }
+    }*/
 
     let mut buffered_msgs: std::vec::Vec<DltMessage> = std::vec::Vec::new();
     let mut buffered_lcs: std::collections::HashSet<LifecycleId> = std::collections::HashSet::new();
@@ -469,7 +504,7 @@ where
                                     while buffered_msgs.len() > 0 {
                                         // todo really inefficient!!!! use drain???
                                         let msg = buffered_msgs.remove(0);
-                                        println!("sending early buffered_msg {:?}", msg);
+                                        // println!("sending early buffered_msg {:?}", msg);
                                         outflow.send(msg).unwrap();
                                     }
                                     assert_eq!(buffered_msgs.len(), 0);
@@ -495,7 +530,7 @@ where
                                 (msg.reception_time_us - msg.timestamp_us()) - max_buffering_delay;
                             // if this is not earlier than the prev_lc start-time we can conclude its a new lifecycle
                             if min_lc_start_time > prev_lc.start_time {
-                                println!("confirmed buffered lc {} as min_lc_start_time {} > prev_lc.start_time {}, lc={:?}", lc2.id, min_lc_start_time, prev_lc.start_time, lc2);
+                                // println!("confirmed buffered lc {} as min_lc_start_time {} > prev_lc.start_time {}, lc={:?}", lc2.id, min_lc_start_time, prev_lc.start_time, lc2);
                                 buffered_lcs.remove(&lc2.id);
                                 lcs_w.insert(lc2.id, new_lifecycle_item(*lc2));
                                 lcs_w.refresh();
@@ -504,7 +539,7 @@ where
                                     while buffered_msgs.len() > 0 {
                                         // todo really inefficient!!!! use drain???
                                         let msg = buffered_msgs.remove(0);
-                                        println!("sending early buffered_msg {:?}", msg);
+                                        // println!("sending early buffered_msg {:?}", msg);
                                         outflow.send(msg).unwrap();
                                     }
                                     assert_eq!(buffered_msgs.len(), 0);
@@ -539,7 +574,7 @@ where
                     // possible criteria:
                     // 1. once the new lifecycle overlaps the prev. one -> and needs a merge (see above)
                     // 2. once the new lifecycle contains messages with timestamp > x (max buffering at (start plus at runtime) buffer)
-                    println!("added lc id {} to buffered_lcs", lc3.id);
+                    // println!("added lc id {} to buffered_lcs", lc3.id);
                     buffered_lcs.insert(lc3.id);
                     ecu_lcs.push(lc3); // we (sadly?) have copy trait for lifecycle
                                        /*
@@ -576,19 +611,19 @@ where
         if buffered_lcs.len() > 0 {
             buffered_msgs.push(msg);
         } else {
-            println!("sending non-buffered_msg {:?}", msg);
+            // todo slog... println!("sending non-buffered_msg {:?}", msg);
             outflow.send(msg).unwrap(); // todo how handle errors?
         }
     }
 
     // if we have still buffered lcs we have to make them valid now:
-    println!("adding {} buffered_lcs to lcs_w at end", buffered_lcs.len());
+    //println!("adding {} buffered_lcs to lcs_w at end", buffered_lcs.len());
     for lc_id in buffered_lcs {
         'outer: for vs in ecu_map.values() {
             for v in vs {
                 if v.id == lc_id {
                     lcs_w.insert(lc_id, new_lifecycle_item(*v));
-                    println!("lcs_w content added at end id={:?} lc={:?}", lc_id, *v);
+                    // println!("lcs_w content added at end id={:?} lc={:?}", lc_id, *v);
                     break 'outer;
                 }
             }
@@ -598,10 +633,11 @@ where
 
     // if we have buffered msgs we have to output them now:
     for m in buffered_msgs.into_iter() {
-        println!("sending buffered_msg {:?}", m);
+        // println!("sending buffered_msg {:?}", m);
         outflow.send(m).unwrap();
     }
 
+    /*
     println!(
         "After processing stream: Have ecu_map.len={}",
         ecu_map.len()
@@ -615,7 +651,7 @@ where
         for (id, b) in a {
             println!("lcs_w content id={:?} lc={:?}", id, b);
         }
-    }
+    }*/
     lcs_w
 }
 
@@ -664,6 +700,28 @@ pub fn get_mapped_lifecycles_as_hashmap<'a>(
         .collect()
 }
 
+/// return a vector of lifecycles sorted by start_time
+/// asserts if an interims lifecycle is contained!
+/// todo add example
+pub fn get_sorted_lifecycles_as_vec<'a, M, S>(
+    lcr: &'a evmap::MapReadRef<LifecycleId, LifecycleItem, M, S>,
+) -> std::vec::Vec<&'a Lifecycle>
+where
+    S: std::hash::BuildHasher + Clone,
+    M: 'static + Clone,
+{
+    let mut sorted_lcs: std::vec::Vec<&'a Lifecycle> = lcr
+        .iter()
+        .map(|(id, b)| {
+            let lc = b.get_one().unwrap();
+            assert_eq!(&lc.id, id);
+            lc
+        })
+        .collect();
+    sorted_lcs.sort_by(|a, b| a.start_time.cmp(&b.start_time));
+    sorted_lcs
+}
+
 #[cfg(test)]
 mod tests {
     //use super::*;
@@ -794,12 +852,17 @@ mod tests {
                     let timestamp_us = start_delay + ((i as u64) * options.frequency); // frequency
                     let min_send_time = std::cmp::max(buf_delay + (i as u64), timestamp_us);
                     msgs.push(DltMessage {
+                        index: i as crate::dlt::DltMessageIndexType,
                         reception_time_us: lc_start_time + min_send_time,
-                        timestamp_dms: (timestamp_us/100)as u32,
+                        timestamp_dms: (timestamp_us / 100) as u32,
                         lifecycle: 0,
                         ecu: options.ecu,
-                        standard_header: crate::dlt::DltStandardHeader{htyp: 1, len: 0, mcnt:0},
-                        extended_header:None,
+                        standard_header: crate::dlt::DltStandardHeader {
+                            htyp: 1,
+                            len: 0,
+                            mcnt: 0,
+                        },
+                        extended_header: None,
                         payload: [].to_vec(),
                     });
                 }
@@ -1069,7 +1132,8 @@ mod tests {
     }
     impl std::cmp::PartialEq for SortedDltMessage {
         fn eq(&self, other: &Self) -> bool {
-            self.lc_start_time + self.m.timestamp_us() == other.lc_start_time + other.m.timestamp_us()
+            self.lc_start_time + self.m.timestamp_us()
+                == other.lc_start_time + other.m.timestamp_us()
         }
     }
     impl std::cmp::Ord for SortedDltMessage {
